/* This is a generated file of CSS imports */
/* It was generated by @jupyterlab/buildutils in Build.ensureAssets() */
@import url('~@jupyterlab/application-extension/style/index.css');
@import url('~@jupyterlab/application/style/index.css');
@import url('~@jupyterlab/apputils-extension/style/index.css');
@import url('~@jupyterlab/codemirror-extension/style/index.css');
@import url('~@jupyterlab/completer-extension/style/index.css');
@import url('~@jupyterlab/console-extension/style/index.css');
@import url('~@jupyterlab/csvviewer-extension/style/index.css');
@import url('~@jupyterlab/docmanager-extension/style/index.css');
@import url('~@jupyterlab/filebrowser-extension/style/index.css');
@import url('~@jupyterlab/fileeditor-extension/style/index.css');
@import url('~@jupyterlab/help-extension/style/index.css');
@import url('~@jupyterlab/imageviewer-extension/style/index.css');
@import url('~@jupyterlab/inspector-extension/style/index.css');
@import url('~@jupyterlab/launcher-extension/style/index.css');
@import url('~@jupyterlab/mainmenu-extension/style/index.css');
@import url('~@jupyterlab/markdownviewer-extension/style/index.css');
@import url('~@jupyterlab/mathjax2-extension/style/index.css');
@import url('~@jupyterlab/notebook-extension/style/index.css');
@import url('~@jupyterlab/rendermime-extension/style/index.css');
@import url('~@jupyterlab/running-extension/style/index.css');
@import url('~@jupyterlab/settingeditor-extension/style/index.css');
@import url('~@jupyterlab/statusbar-extension/style/index.css');
@import url('~@jupyterlab/tabmanager-extension/style/index.css');
@import url('~@jupyterlab/terminal-extension/style/index.css');
@import url('~@jupyterlab/tooltip-extension/style/index.css');
@import url('~@jupyterlab/ui-components-extension/style/index.css');
